/*	cmsinit_simple - Initialize GPRS connection
 *
 *	Authors: joshua.randall@orange-ftgroup.com
 *               JIANG Dan <dan.jiang@orange-ftgroup.com> 
 *               YU Yijun     <yijun.yu@orange-ftgroup.com>   
 *
 * Modified from serial.c of GNOKII package
 */

/*

  $Id: serial.c,v 1.1.1.1 2006/09/26 03:17:28 nanye Exp $

  G N O K I I

  A Linux/Unix toolset and driver for the mobile phones.

  This file is part of gnokii.

  Gnokii is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Gnokii is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with gnokii; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Copyright (C) 1999-2000  Hugh Blemings & Pavel Jank ml.
  Copyright (C) 2001       Chris Kemp, Manfred Jonsson, Jank Kratochvil
  Copyright (C) 2002       Ladis Michl, Pavel Machek
  Copyright (C) 2001-2004  Pawel Kot
  Copyright (C) 2002-2004  BORBELY Zoltan
  
*/


/* Do not compile this file under Win32 systems. */

#ifndef WIN32

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
//#include <glib.h>
#include "serial.h"
#include "config.h"
//#include "util.h"


#include <termios.h>

#ifdef HAVE_SYS_IOCTL_COMPAT_H
#  include <sys/ioctl_compat.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#  include <sys/select.h>
#endif

#ifdef HAVE_SYS_MODEM_H
#  include <sys/modem.h>
#endif

#ifdef HAVE_SYS_FILIO_H
#  include <sys/filio.h>
#endif

/* If the target operating system does not have cfsetspeed, we can emulate
   it. */


#ifndef O_NONBLOCK
#  define O_NONBLOCK  0
#endif

/* Structure to backup the setting of the terminal. */
struct termios serial_termios;
int serial_close (int fd);


/* Open the serial port and store the settings. */
int
serial_open (const char *file, int oflag)
{
  int fd;
  int retcode;

  fd = open (file, oflag);
  if (fd == -1) {
    perror ("Gnokii serial_open: open");
    return -1;
  }

  retcode = tcgetattr (fd, &serial_termios);
  if (retcode == -1) {
    perror ("Gnokii serial_open: tcgetattr");
    /* Don't call serial_close since serial_termios is not valid */
    close (fd);
    return -1;
  }

  return fd;
}

/* Close the serial port and restore old settings. */
int
serial_close (Int32 fd)
{
  if (fd >= 0) {
#if 1                           /* HACK */
    serial_termios.c_cflag |= HUPCL;    /* production == 1 */
#else
    serial_termios.c_cflag &= ~HUPCL;   /* debugging  == 0 */
#endif

    tcsetattr (fd, TCSANOW, &serial_termios);
  }

  return close (fd);
}

/* Open a device with standard options.
 * Use value (-1) for "with_hw_handshake" if its specification is required from the user.
 */
int
serial_opendevice (const Int8 * file, Int32 with_odd_parity, Int32 with_async)
{
  int fd;
  int retcode;
  struct termios tp;

  /* Open device */

  /* O_NONBLOCK MUST be used here as the CLOCAL may be currently off
   * and if DCD is down the "open" syscall would be stuck wating for DCD.
   */
  fd = serial_open (file, O_RDWR | O_NOCTTY | O_NONBLOCK);

  if (fd < 0)
    return fd;

  /* Initialise the port settings */
  memcpy (&tp, &serial_termios, sizeof (struct termios));

  /* Set port settings for canonical input processing */
  tp.c_cflag = B0 | CS8 | CLOCAL | CREAD | HUPCL;
  if (with_odd_parity) {
    tp.c_cflag |= (PARENB | PARODD);
    tp.c_iflag = 0;
  } else
    tp.c_iflag = IGNPAR;
#ifdef CRTSCTS
  tp.c_cflag &= ~CRTSCTS;
#endif

  tp.c_oflag = 0;
  tp.c_lflag = 0;
  tp.c_cc[VMIN] = 1;
  tp.c_cc[VTIME] = 0;

  retcode = tcflush (fd, TCIFLUSH);
  if (retcode == -1) {
    perror ("Gnokii serial_opendevice: tcflush");
    serial_close (fd);
    return -1;
  }

  retcode = tcsetattr (fd, TCSANOW, &tp);
  if (retcode == -1) {
    perror ("Gnokii serial_opendevice: tcsetattr");
    serial_close (fd);
    return -1;
  }

  serial_changespeed (fd, 115200);

  /* We need to turn off O_NONBLOCK now (we have CLOCAL set so it is safe).
   * When we run some device script it really doesn't expect NONBLOCK!
   */

  retcode = fcntl (fd, F_SETFL, 0);
  if (retcode == -1) {
    perror ("Gnokii serial_opendevice: fnctl(F_SETFL)");
    serial_close (fd);
    return -1;
  }

  /* Allow process/thread to receive SIGIO */

#if !(__unices__)
  retcode = fcntl (fd, F_SETOWN, getpid ());
  if (retcode == -1) {
    perror ("Gnokii serial_opendevice: fnctl(F_SETOWN)");
    serial_close (fd);
    return -1;
  }
#endif

  /* Make filedescriptor asynchronous. */

  /* We need to supply FNONBLOCK (or O_NONBLOCK) again as it would get reset
   * by F_SETFL as a side-effect!
   */
#ifdef FNONBLOCK
  retcode = fcntl (fd, F_SETFL, (with_async ? FASYNC : 0) | FNONBLOCK);
#else
#  ifdef FASYNC
  retcode = fcntl (fd, F_SETFL, (with_async ? FASYNC : 0) | O_NONBLOCK);
#  else
  retcode = fcntl (fd, F_SETFL, O_NONBLOCK);
  if (retcode != -1)
    retcode = ioctl (fd, FIOASYNC, &with_async);
#  endif
#endif
  if (retcode == -1) {
    perror ("Gnokii serial_opendevice: fnctl(F_SETFL)");
    serial_close (fd);
    return -1;
  }

  return fd;
}

/* Set the DTR and RTS bit of the serial device. */
void
serial_setdtrrts (int fd, int dtr, int rts)
{
  unsigned int flags;

  flags = TIOCM_DTR;

  if (dtr)
    ioctl (fd, TIOCMBIS, &flags);
  else
    ioctl (fd, TIOCMBIC, &flags);

  flags = TIOCM_RTS;

  if (rts)
    ioctl (fd, TIOCMBIS, &flags);
  else
    ioctl (fd, TIOCMBIC, &flags);
}


int
serial_select (int fd, struct timeval *timeout)
{
  fd_set readfds;

  FD_ZERO (&readfds);
  FD_SET (fd, &readfds);

  return select (fd + 1, &readfds, NULL, NULL, timeout);
}

static int
serial_wselect (int fd, struct timeval *timeout)
{
  fd_set writefds;

  FD_ZERO (&writefds);
  FD_SET (fd, &writefds);

  return select (fd + 1, NULL, &writefds, NULL, timeout);
}


/* Change the speed of the serial device.
 * RETURNS: Success
 */
int
serial_changespeed (int fd, int speed)
{
  int retcode = 0;

#ifndef SGTTY
  struct termios t;
#else
  struct sgttyb t;
#endif
  int new_speed = B9600;

  switch (speed) {
    case 9600:
      new_speed = B9600;
      break;
    case 19200:
      new_speed = B19200;
      break;
    case 38400:
      new_speed = B38400;
      break;
    case 57600:
      new_speed = B57600;
      break;
    case 115200:
      new_speed = B115200;
      break;
    default:
      fprintf (stderr, "Serial port speed %d not supported!\n", speed);
      return -1;
  }

#ifndef SGTTY
  if (tcgetattr (fd, &t))
    retcode = 1;

  if (cfsetspeed (&t, new_speed) == -1) {
    printf ("Serial port speed setting failed\n");
    retcode = 2;
  }

  tcsetattr (fd, TCSADRAIN, &t);
#else
  if (ioctl (fd, TIOCGETP, &t))
    retcode = 3;

  t.sg_ispeed = new_speed;
  t.sg_ospeed = new_speed;

  if (ioctl (fd, TIOCSETN, &t))
    retcode = 4;
#endif
  return retcode;
}

/* Read from serial device. */
size_t
serial_read (Int32 fd, void *buf, size_t nbytes)
{
  return read (fd, buf, nbytes);
}

#if !defined(TIOCMGET) && defined(TIOCMODG)
#  define TIOCMGET TIOCMODG
#endif


/* Write to serial device. */
size_t
serial_write (Int32 fd, const Int8 * buf, size_t n, Int32 serial_write_usleep)
{
  size_t r = 0, bs;
  ssize_t got;

  while (n > 0) {
    bs = n;
    got = write (fd, buf + r, bs);
    if (got == 0) {
      printf ("Serial write: oops, zero byte has written!\n");
    } else if (got < 0) {
      if (errno == EINTR)
        continue;
      if (errno != EAGAIN) {
        printf ("Serial write: write error %d\n", errno);
        return -1;
      }
      printf ("Serial write: transmitter busy, waiting\n");
      serial_wselect (fd, NULL);
      printf ("Serial write: transmitter ready\n");
      continue;
    }

    n -= got;
    r += got;
  }
  return r;
}

int
serial_nreceived (int fd, int *n)
{
  if (ioctl (fd, FIONREAD, n)) {
    printf ("serial_nreceived: cannot get the received data size\n");
    return -1;
  }

  return 0;
}

int
serial_flush (int fd)
{
  if (tcdrain (fd)) {
    printf ("serial_flush: cannot flush serial device\n");
    return -1;
  }

  return 0;
}

#endif /* WIN32 */
